#ifndef SL_COMPLEX_CCALC_H
#define SL_COMPLEX_CCALC_H

/*
 ******************************
 * Scientific Library (GNU Public Licence)
 *
 * Author: Laurent Deniau, Laurent.Deniau@cern.ch
 *
 * $Id: ccalc.h,v 1.6 1998/07/09 07:05:50 paubert Exp $
 *
 * Suggestions: sl@mathinsa.insa-lyon.fr
 * Bugs:   sl-bugs@mathinsa.insa-lyon.fr
 *
 * For more information, please see the sl++ Home Page:
 * http://wwwinfo.cern.ch/~ldeniau/sl.html
 *
 ******************************
 */

#ifndef SL_COMPLEX_H
#error <sl/complex/ccalc.h> must be included via <sl/complex.h>
#endif

#ifdef HAVE_NAMESPACE
namespace sl
{
#endif

// Power/Logarithmic functions

template <typename T>
inline Complex<T>
ln (Complex<T> const& z)
{
  return Complex<T> (log(abs(z)), arg(z));
}

template <typename T>
inline Complex<T>
log (Complex<T> const& z)
{
  return Complex<T> (log(abs(z)), arg(z));
}

template <typename T>
inline Complex<T>
log10 (Complex<T> const& z)
{
  return log(z)/T(M_LN10);
}

template <typename T>
inline Complex<T>
exp (Complex<T> const& z)
{
  return rect( T(exp(re(z))), im(z) );
}

template <typename T1, typename T2>
inline typename promote<Complex<T1>,Complex<T2> >::promoted_t
pow (Complex<T1> const& z1, Complex<T2> const& z2)
{
  typedef typename promote<T1,T2>::promoted_t promoted_t;
  promoted_t r = promoted_t(log(abs(z1)));
  promoted_t t = promoted_t(arg(z1));
  return rect( promoted_t(exp(r*re(z2) - im(z2)*t)), promoted_t(im(z2)*r + re(z2)*t) );
}

template <typename T>
inline Complex<T>
pow (Complex<T> const& z, T const x)
{
  return exp(x*ln(z));
}

template <typename T>
inline Complex<T>
pow (T const x, Complex<T> const& z)
{
  return exp(z*T(log(x)));
}

template <typename T>
inline Complex<T>
pow10 (Complex<T> const& z)
{
  return exp(z*T(M_LN10));
}

template <typename T>
inline Complex<T>
sqrt (Complex<T> const& z)
{
  T n = T(abs(z));
  T nr, ni;
  if (n == 0) nr = ni = n;
  else if (re(z) >= 0) {
    nr = sqrt((n+re(z))*0.5);
    ni = im(z)*0.5/nr;
  } else {
    ni = T(sqrt((n-re(z))*0.5));
    if (im(z) < 0) ni = -ni;
    nr = im(z)*0.5/ni;
  }
  return Complex<T> (nr, ni); 
}

// Trigonometric functions

template <typename T>
inline Complex<T>
sin (Complex<T> const& z)
{
  return Complex<T> (sin(re(z))*cosh(im(z)), cos(re(z))*sinh(im(z)));
}

template <typename T>
inline Complex<T>
cos (Complex<T> const& z)
{
  return Complex<T> (cos(re(z))*cosh(im(z)), -sin(re(z))*sinh(im(z)));
}

template <typename T>
inline Complex<T>
tan (Complex<T> const& z)
{
  return Complex<T> ( sin(z)/cos(z) );
}

// Hyperbolic functions

template <typename T>
inline Complex<T>
sinh (Complex<T> const& z)
{
  return Complex<T> (sinh(re(z))*cos(im(z)), cosh(re(z))*sin(im(z)));
}

template <typename T>
inline Complex<T>
cosh (Complex<T> const& z)
{
  return Complex<T> (cosh(re(z))*cos(im(z)), sinh(re(z))*sin(im(z)));
}

template <typename T>
inline Complex<T>
tanh (Complex<T> const& z)
{
  return Complex<T> ( sinh(z)/cosh(z) );
}

// Reciprocal hyperbolic functions

template <typename T>
inline Complex<T>
asinh (Complex<T> const& z)
{
  return -ln(sqrt(T(1)+z*z)-z);
}

template <typename T>
inline Complex<T>
acosh (Complex<T> const& z)
{
  return ln(sqrt(z*z-T(1))+z);
}

template <typename T>
inline Complex<T>
atanh (Complex<T> const& z)
{
  Complex<T> const I(0,1);
  Complex<T> Ix = exri(conj(z));
  return ln((I+Ix)/(I-Ix))/T(2);
}

// Reciprocal trigonometric functions

template <typename T>
inline Complex<T>
asin (Complex<T> const& z)
{
  return conj(exri(ln(sqrt(T(1)-z*z)+exri(conj(z)))));
}

template <typename T>
inline Complex<T>
acos (Complex<T> const& z)
{
  if (re(z)*im(z) < 0)
    return exri(conj(acosh(z)));
  else
    return conj(exri(acosh(z)));
}

template <typename T>
inline Complex<T>
atan (Complex<T> const& z)
{
    Complex<T> const I(0,1);
    return exri(conj(ln((I+z)/(I-z))/T(2)));
}

#ifdef HAVE_NAMESPACE
}
#endif

#endif // SL_COMPLEX_CCALC_H
