#ifndef SL_COMPLEX_COMPLEX_H
#define SL_COMPLEX_COMPLEX_H

/*
 ******************************
 * Scientific Library (GNU Public Licence)
 *
 * Author: Laurent Deniau, Laurent.Deniau@cern.ch
 *
 * $Id: complex.h,v 1.7 1998/07/08 13:25:36 paubert Exp $
 *
 * Suggestions: sl@mathinsa.insa-lyon.fr
 * Bugs:   sl-bugs@mathinsa.insa-lyon.fr
 *
 * For more information, please see the sl++ Home Page:
 * http://wwwinfo.cern.ch/~ldeniau/sl.html
 *
 ******************************
 */

#ifndef SL_COMPLEX_H
#error <sl/complex/complex.h> must be included via <sl/complex.h>
#endif

/*
 * Remark: I will use of numeric_limits<T>::epsilon() instead of DBL_EPSILON
 * when the std header <limits> will be compatible with C++ ANSI
 */

#ifdef HAVE_NAMESPACE
namespace sl 
{
#endif




    template <typename T =Real>
        class Complex {


            public:

            typedef T value_t;


            public:

            T real() const { return my_re; }
            T imag() const { return my_im; }


            public:

              // TODO: For efficiency the constructor must be explicit
              // changes will be made when egcs will support template overload...

            Complex<T> (T const x=T(), T const y=T())
            : my_re( x ), my_im( y ) {
            }

// The compiler will display a warning if a lost of precision occurs
// It is possible to use checked_cast<T,T2>(z.real()) if we want a runtime
// error for sizeof(T)<sizeof(T2)...

            template <typename T2>
            Complex (Complex<T2> const& z) 
            : my_re( z.real() ), my_im( z.imag() ) {
            }


            public:

            operator Bool () const {
                return my_re && my_im;
            }

            Complex<T>&
            operator = (T const x) {
                my_re = x;
                my_im = T();
                return *this;
            }

            Complex<T>&
            operator += (T const x) {
                my_re += x;
                return *this;
            }

            Complex<T>&
            operator -= (T const x) {
                my_re -= x;
                return *this;
            }

            Complex<T>&
            operator *= (T const x) {
                my_re *= x;
                my_im *= x;
                return *this;
            }

            Complex<T>&
            operator /= (T const x) {
                my_re /= x;
                my_im /= x;
                return *this;
            }

            template <typename T2>
            Complex<T>&
            operator = (Complex<T2> const& z) {
                my_re = z.real();
                my_im = z.imag();
                return *this;
            }

            template <typename T2>
            Complex<T>&
            operator += (Complex<T2> const& z) {
                my_re += z.real();
                my_im += z.imag();
                return *this;
            }

            template <typename T2>
            Complex<T>&
            operator -= (Complex<T2> const& z) {
                my_re -= z.real();
                my_im -= z.imag();
                return *this;
            }

            template <typename T2>
            Complex<T>&
            operator *= (Complex<T2> const& z) {
                T t   = my_re * z.real() - my_im * z.imag();
                my_im = my_re * z.imag() + my_im * z.real();
                my_re = t;
                return *this;
            }

            template <typename T2>
            Complex<T>&
            operator /= (Complex<T2> const& z) {
                typedef typename promote<T,T2>::promoted_t promoted_t;

                T nr, ni;
                if (abs(z.real()) <= abs(z.imag())) {
                    promoted_t t = promoted_t(z.real())/z.imag();
                    promoted_t n = z.imag()*(1+t*t);
                    nr = (my_re*t+my_im)/n;
                    ni = (my_im*t-my_re)/n;
                } else {
                    promoted_t t = promoted_t(z.imag())/z.real();
                    promoted_t n = z.real()*(1+t*t);
                    nr = (my_re+my_im*t)/n;
                    ni = (my_im-my_re*t)/n;
                }
                my_re = nr;
                my_im = ni;
                return *this;
            }

            private:
            T my_re, my_im;
        };

#ifdef HAVE_NAMESPACE
}
#endif 

#endif // SL_COMPLEX_COMPLEX_H
